### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2009 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

__docformat__ = "restructuredtext"

# import standard packages

# import Zope3 interfaces
from zope.app.appsetup.interfaces import IDatabaseOpenedWithRootEvent
from zope.app.catalog.interfaces import ICatalog
from zope.app.intid.interfaces import IIntIds
from zope.app.publication.zopepublication import ZopePublication
from zope.component.interfaces import IComponentRegistry
from zope.dublincore.interfaces import IZopeDublinCore
from zope.location.interfaces import ISite

# import local interfaces
from hurry.workflow.interfaces import IWorkflowState
from ztfy.utils.interfaces import INewSiteManagerEvent

# import Zope3 packages
from zc.catalog.catalogindex import ValueIndex, DateTimeValueIndex
from zope.app.catalog.catalog import Catalog
from zope.app.component import hooks
from zope.app.generations.utility import findObjectsProviding
from zope.app.intid import IntIds
from zope.component import adapter
from zope.location import locate

# import local packages
from ztfy.utils.site import locateAndRegister


def updateDatabaseIfNeeded(context, create_catalog=False):
    """Check for missing objects at application startup"""
    try:
        sm = context.getSiteManager()
    except:
        return
    default = sm['default']
    # Check for required IIntIds utility
    intids = default.get('IntIds')
    if intids is None:
        intids = IntIds()
        locate(intids, default)
        IComponentRegistry(sm).registerUtility(intids, IIntIds, '')
        default['IntIds'] = intids
    # Check for required catalog and index
    catalog = default.get('WorkflowCatalog')
    if (catalog is None) and create_catalog:
        catalog = Catalog()
        locateAndRegister(catalog, default, 'WorkflowCatalog', intids)
        IComponentRegistry(sm).registerUtility(catalog, ICatalog, 'WorkflowCatalog')
    if catalog is not None:
        if 'wf_id' not in catalog:
            index = ValueIndex('getId', IWorkflowState, True)
            locateAndRegister(index, catalog, 'wf_id', intids)
        if 'wf_state' not in catalog:
            index = ValueIndex('getState', IWorkflowState, True)
            locateAndRegister(index, catalog, 'wf_state', intids)
        if 'creation_date' not in catalog:
            index = DateTimeValueIndex('created', IZopeDublinCore, False)
            locateAndRegister(index, catalog, 'creation_date', intids)
        if 'modification_date' not in catalog:
            index = DateTimeValueIndex('modified', IZopeDublinCore, False)
            locateAndRegister(index, catalog, 'modification_date', intids)
        if 'effective_date' not in catalog:
            index = DateTimeValueIndex('effective', IZopeDublinCore, False)
            locateAndRegister(index, catalog, 'effective_date', intids)
        if 'expiration_date' not in catalog:
            index = DateTimeValueIndex('expires', IZopeDublinCore, False)
            locateAndRegister(index, catalog, 'expiration_date', intids)


@adapter(IDatabaseOpenedWithRootEvent)
def handleOpenedDatabase(event):
    db = event.database
    connection = db.open()
    root = connection.root()
    root_folder = root.get(ZopePublication.root_name, None)
    updated = False
    for site in findObjectsProviding(root_folder, ISite):
        if site is not root_folder:
            hooks.setSite(site)
            updateDatabaseIfNeeded(site)
            updated = True
    if not updated:
        hooks.setSite(root_folder)
        updateDatabaseIfNeeded(root_folder)


@adapter(INewSiteManagerEvent)
def handleNewSiteManager(event):
    updateDatabaseIfNeeded(event.object, create_catalog=True)



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2009 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


# import Zope3 interfaces
from zope.schema.interfaces import IVocabularyTokenized

# import local interfaces
from hurry.workflow.interfaces import IWorkflow as IWorkflowBase
from ztfy.comment.interfaces import ICommentable

# import Zope3 packages
from zope.interface import Interface, invariant, Invalid
from zope.schema import Datetime, Choice, Tuple, TextLine, Object

# import local packages
from ztfy.security.schema import Principal

from ztfy.workflow import _


class IWorkflow(IWorkflowBase):
    """Marker interface for custom workflows extending IWorkflow"""

    states = Object(schema=IVocabularyTokenized)

    published_states = Tuple(title=_("Published states"),
                             description=_("Tuple of published and potentially visible workflow states"),
                             required=False,
                             value_type=TextLine())


class IWorkflowTarget(ICommentable):
    """Marker interface for contents handled by a workflow"""

    workflow_name = Choice(title=_("Workflow name"),
                           description=_("Name of a registered workflow utility used by this workflow"),
                           required=True,
                           vocabulary="ZTFY workflows")


class IWorkflowContentInfo(Interface):
    """Interface used to define common workflow properties"""

    state_date = Datetime(title=_("State date"),
                          description=_("Date when the current state was defined"))

    state_principal = Principal(title=_("State principal"),
                                description=_("Name of the principal who defined the current state"))

    publication_date = Datetime(title=_("Publication date"),
                                description=_("Date when content was accepted to publication by the publisher, entered in ISO format"),
                                required=False)

    first_publication_date = Datetime(title=_("First publication date"),
                                      description=_("Date when content was accepted to publication by the publisher for the first time"),
                                      required=False)

    publication_effective_date = Datetime(title=_("Publication start date and time"),
                                          description=_("Date and time from which content will be visible"),
                                          required=True,
                                          default=None)

    publication_expiration_date = Datetime(title=_("Publication end date and time"),
                                           description=_("Date and time until which content will be visible"),
                                           required=False,
                                           default=None)

    @invariant
    def expireAfterEffective(self):
        if self.publication_expiration_date is not None:
            if self.publication_effective_date is None:
                raise Invalid(_("Can't define publication end date without publication start date"))
            if self.publication_expiration_date <= self.publication_effective_date:
                raise Invalid(_("Publication end date must be null or defined after publication's start date"))

    def isPublished():
        """Is the context published or not ?"""

    def isVisible():
        """Is the context visible according to current workflow state and security context ?"""


class IWorkflowContent(IWorkflowContentInfo):
    """Marker interface for workflow contents"""



### -*- coding: utf-8 -*- ####################################################
##############################################################################
#
# Copyright (c) 2008-2009 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


# import Zope3 interfaces

# import local interfaces
from hurry.query.interfaces import IQuery
from hurry.workflow.interfaces import IWorkflowVersions, IWorkflowState

# import Zope3 packages
from zope.app import zapi
from zope.interface import implements

# import local packages
from hurry.query import And
from hurry.query.value import In, Eq
from hurry.workflow.workflow import WorkflowVersions as WorkflowVersionsBase


WF_STATE_INDEX = ('WorkflowCatalog', 'wf_state')
WF_IDS_INDEX = ('WorkflowCatalog', 'wf_id')


class WorkflowVersions(WorkflowVersionsBase):
    """Utility used to handle content versions"""

    implements(IWorkflowVersions)

    def getVersions(self, state=None, id=None, object=None):
        assert (state is not None) or (id is not None) or (object is not None)
        query = zapi.getUtility(IQuery)
        request = []
        if state is not None:
            if isinstance(state, tuple):
                request.append(In(WF_STATE_INDEX, state))
            else:
                request.append(Eq(WF_STATE_INDEX, state))
        if id is not None:
            request.append(Eq(WF_IDS_INDEX, id))
        elif object is not None:
            state = IWorkflowState(object, None)
            if state is not None:
                request.append(Eq(WF_IDS_INDEX, state.getId()))
        return query.searchResults(And(*request))

    def getVersionsWithAutomaticTransitions(self):
        return ()

    def hasVersion(self, state, id):
        return bool(self.getVersions(state, id))

    def hasVersionId(self, id):
        query = zapi.getUtility(IQuery)
        return bool(query.searchResults(Eq(WF_IDS_INDEX, id)))
